/* ==================================================================
   File:        kscore.cpp
   Author:      Steve Monks
   Description:
	Kernel Streaming playback manager

 	$Id: kscore.cpp,v 1.10 2006/04/09 19:51:53 Steve Exp $
	$Log: kscore.cpp,v $
	Revision 1.10  2006/04/09 19:51:53  Steve
	Version 3.6
	Added logarithmic volume control
	Added diagnostic logging
	
	Revision 1.9  2006/03/12 15:57:09  Steve
	Fixed incorrect play reporting issue
	
	Revision 1.8  2006/02/11 14:06:00  Steve
	Fix for Gapless Mode crashes when format changes.
	Fix for Gapless Mode lockup when track changed while paused
	Added format info to status window
	Fixes to Release CFG project setting (removed debug info and _DEBUG define)
	
	Revision 1.7  2006/02/04 21:57:38  Steve
	Added out shutdown of KS when buffer underruns in gapless mode
	tweaks to readme
	
	Revision 1.6  2006/02/04 20:59:44  Steve
	Initial import of readme.html
	
	Added gapless mode (for evaluation purposes)
	
	Revision 1.5  2006/01/30 21:07:29  Steve
	Added protection against daft values to config stuff
	Fixed config dialog so that KSCore is only recreated when settings actually change
	Changed buffer size readouts to KB in config dialog
	Fixed potential problem with lost volume settings when KScore is deleted and recreated
	
	Revision 1.4  2006/01/29 21:33:57  Steve
	Added an adjustment for the number of output buffers
	Removed buffer size radio buttons and replaced with sliders
	Now deletes and recreates the KSCore object when a new track starts after changing the settings
	
	Revision 1.3  2006/01/08 22:44:37  Steve
	Added status window
	Added increased buffer settings
	Fixed .d00 file problems (returned truthful buffer sizes rather than 8192)
	
	Revision 1.2  2006/01/06 22:26:20  Steve
	Fixed problem with Winamp Music and .mod playback (checking against packet_size
	rather than len inside Write function).
	Removed delay before player thread sleeps when starved of data
	
	Revision 1.1  2006/01/03 21:46:13  Steve
	Initial import of V3.0.
	
	This is a complete rewrite of the plugin and Kernel streaming code.
	The only code retained from the previous 2.x version is the Microsoft
	sample code and the filter enumeration.
	
	
================================================================== */

#include "kssample.h"
#include "stdarg.h"
#include "error.h"
#include "deviceinfo.h"
#include "config.h"
#include "kscore.h"
#include "resource.h"

static KSCore *kscore = NULL;

void KSCore::FlushInputBuffer(void)
{
	read_ptr = write_ptr = input_buffer;
	input_buffer_len = 0;
}

// returns TRUE if pkt is filled

int KSCore::CopyFromInput(int pkt, int amt)
{
	char *dst;
	int acnt, bcnt, pkt_free;

	pkt_free = packet_size-packets[cur_packet].Header.DataUsed;	// amount free in current packet
	amt = min(amt,pkt_free);												// amount to copy
	
	dst = (char *)packets[pkt].Header.Data + packets[pkt].Header.DataUsed;
	acnt = min(amt,input_buffer_size - (read_ptr-input_buffer));	// smallest of amount to copy or space to end of buffer
	bcnt = 0;
	if (acnt < amt) {
		bcnt = amt-acnt;
	}
	if (acnt) {
		CopyMemoryAttenuated(dst,(void *)read_ptr,acnt);
		read_ptr += acnt;
		dst += acnt;
	}
	if (bcnt) {													// deal with wrapped block
		read_ptr = input_buffer;
		CopyMemoryAttenuated(dst,(void *)read_ptr,bcnt);
		read_ptr += bcnt;
		dst += bcnt;
	}
	packets[pkt].Header.DataUsed += amt;				// update data count in packet
	input_buffer_len -= amt;									// update data count in temp buffer

	if (packets[pkt].Header.DataUsed == packet_size)
		return TRUE;											// packet is full
	else
		return FALSE;											// packet is not full
}



void KSCore::CopyMemoryAttenuated(void *dest, void *source, int len)
{
	typedef union {
		char *c;
		short *s;
		int *i;
		float *f;
		void *v;
	} VP;

	VP d, s;
	union {
		int i;
		char c[4];
	} v;

	if (volume != 255 && config->volume_enabled) {
		switch(wfx_cur.Format.wBitsPerSample)
		{
		case 8:						// 8 bit samples
			d.v = dest;
			s.v = source;
			while(len >= 1) {
				v.i = *s.c;
				v.i = (v.i * volume) >> 8;
				*d.c = v.i;
				++d.c;
				++s.c;
				len -= 1;
			}
			return;

		case 16:						// 16 bit samples
			d.v = dest;
			s.v = source;
			while(len >= 2) {
				v.i = *s.s;
				v.i = (v.i * volume) >> 8;
				*d.s = v.i;
				++d.s;
				++s.s;
				len -= 2;
			}
			return;

		case 32:						// 24 bit samples in a 32 bit package
			d.v = dest;
			s.v = source;
			v.i = 0;
			while(len >= 4) {
				v.i = s.i[0];	// read 24 bit sample embedded in top 3 bytes of 32 bit data
				v.i >>= 8;
				v.i *= volume;
				d.i[0] = v.i;
				++d.i;
				++s.i;
				len -= 4;

			}
			return;

		default:
			// unsupported format, just memory copy it by falling through
			break;
		}	
	}
	// unsupported format, just copy the data unmodified
	CopyMemory(dest,source, len);
}


void KSCore::WakePlayer(void)
{
	EnterCriticalSection(&cs_msg);
	last_tick = timeGetTime()&0x7fffffff;
	if (WaitForSingleObject(sleep_event, 0) == WAIT_TIMEOUT)
		DebugMsgNoLog("Thread woke at %d",last_tick);

	SetEvent(sleep_event);
	LeaveCriticalSection(&cs_msg);
}


DWORD WINAPI KSCore::PlayerThread(LPVOID parm)
{
	
	int ix,len,full;
	HRESULT hr;
	int cur_time;

	do {
		// stall here if we've been put to sleep
		WaitForSingleObject(sleep_event, INFINITE);
		dbg_player_sleeps = FALSE;	// thread is awake!

		// sit here waiting for something to happen
		ix = WaitForMultipleObjects(EV_KSPACKET+num_packets, events, FALSE, INFINITE);
		ix -= WAIT_OBJECT_0;


		switch(ix)
		{
		case EV_MSG:
			ResetEvent(events[EV_MSG]);
			break;

		default:	// a ks packet has just come free
			if (cur_packet < 0) {
				cur_packet = ix-EV_KSPACKET;
				packets[cur_packet].Header.DataUsed = 0;					// empty packet

			}
			break;
		}

		len = input_buffer_len;							// copy to local variable as it may change (can only get bigger though)

		if (len && cur_packet >= 0 && !paused) {	// going to write to output buffer
			full = CopyFromInput(cur_packet,len);
			if (full || mode == MODE_STOPPING) {		// packet was filled, needs to be sent
				packets[cur_packet].Header.FrameExtent = packets[cur_packet].Header.DataUsed;
				written_data += packets[cur_packet].Header.DataUsed;
				ResetEvent(packets[cur_packet].Signal.hEvent);
				hr = pin->WriteData(&packets[cur_packet].Header, &packets[cur_packet].Signal);
				if (!SUCCEEDED(hr))
					DebugMsg("KS Packet write failed (buffer %d)",cur_packet);
				cur_packet = -1;	// need to get another ks packet
			}
		}
		else {												// nothing to write! Suspend thread
			EnterCriticalSection(&cs_msg);
			cur_time = timeGetTime()&0x7fffffff;

			ResetEvent(sleep_event);	// go to sleep
			DebugMsgNoLog("Thread sleeps at %d",cur_time);
			dbg_player_sleeps = TRUE;
			if (mode != MODE_STOPPING && !should_be_playing && config->gapless_enabled) {	// buffer is empty and we're in gapless mode?
				DebugMsg("Gapless mode buffer underrun - Player Thread is stopping KS player");
				// this should be pretty safe as it will only do this if the
				// buffer underruns and Winamp says we should have stopped
				// so the only situation under which we'll get a false End
				// is if Start is invoked on the main thread just as we enter
				// into this End which will stop it again as soon as Start
				// has finished its stuff.
				End(TRUE);
			}

			LeaveCriticalSection(&cs_msg);
		}
		thread_paused = paused;

	} while(!kscore->quit);

	DebugMsg("Player Thread - closing");
	player_thread = NULL;
	return 0;
}

DWORD WINAPI KSCore::PlayerThreadWrapper(LPVOID parm)
{
	return kscore->PlayerThread(parm);
}



CKsEnumFilters *KSCore::OpenEnumerator(void)
{
	HRESULT hr;

	if (!the_enumerator)
	{

		// enumerate audio renderers
	   the_enumerator = new CKsEnumFilters(&hr);

		if(the_enumerator == NULL)
		{
			MessageBox(hwnd, "Failed to allocate CKsEnumFilters", "Error!", MB_OK);
			return NULL;
		}
		else if(!SUCCEEDED(the_enumerator))
		{
			MessageBox(hwnd, "Failed to create CKsEnumFilters", "Error!", MB_OK);
			return NULL;
		}

	    GUID  aguidEnumCats[] = { STATIC_KSCATEGORY_AUDIO, STATIC_KSCATEGORY_RENDER };

	    hr = 
	        the_enumerator->EnumFilters
	        (
	            eAudRen,            // create audio render filters ...
	            aguidEnumCats,      // ... of these categories
	            2,                  // There are 2 categories
	            TRUE,               // While you're at it, enumerate the pins
	            FALSE,              // ... but don't bother with nodes
	            TRUE                // Instantiate the filters
	        );

	    if(!SUCCEEDED(hr))
		{
			MessageBox(hwnd, "CKsEnumFilters::EnumFilters failed", "Error!", MB_OK);
			return NULL;
		}
	}
	return the_enumerator;
}



void KSCore::CloseEnumerator(void)
{
	if (the_enumerator) {
		delete the_enumerator;
		the_enumerator = NULL;
	}
}


CKsAudRenFilter *KSCore::PickFilter(void)
{
	CKsAudRenFilter	*filter;
	CKsEnumFilters		*pEnumerator;

	pEnumerator = the_enumerator;

	if (!pEnumerator)
		return NULL;


	pEnumerator->m_listFilters.GetHead((CKsFilter**)&filter);
	{
		LISTPOS lp = pEnumerator->m_listFilters.GetHeadPosition();
		filter = NULL;

		// Pick output device by matching name with current selection
		do {
			pEnumerator->m_listFilters.GetNext(lp, (CKsFilter**)&filter);	// Wave1
			if (filter) {
				if (0 == stricmp(config->out_name,filter->GetName()))
					break;
				filter = NULL;
			}
			else break;
		} while(1);

		if (!filter) {	// no match, pick first filter
			lp = pEnumerator->m_listFilters.GetHeadPosition();
			pEnumerator->m_listFilters.GetNext(lp, (CKsFilter**)&filter);	// Wave1
		}
	}

	return filter;
}


void KSCore::ConvertWaveFormatExtensibleToWaveFormat(WAVEFORMATEXTENSIBLE *out, WAVEFORMATEXTENSIBLE *in)
{
	*out = *in;

	if (out->Samples.wValidBitsPerSample == 24) {
		out->Format.wBitsPerSample = 24;
		if (out->Format.nChannels == 2) {
			out->Format.nBlockAlign    = 6;
		}
		else {
			out->Format.nBlockAlign    = 3;
		}
		out->Format.nAvgBytesPerSec = out->Format.nSamplesPerSec * out->Format.nBlockAlign;
	}
	else if (out->Samples.wValidBitsPerSample == 32) {
		out->Format.wBitsPerSample = 32;
		if (out->Format.nChannels == 2) {
			out->Format.nBlockAlign    = 8;
		}
		else {
			out->Format.nBlockAlign    = 4;
		}
		out->Format.nAvgBytesPerSec = out->Format.nSamplesPerSec * out->Format.nBlockAlign;
	}

	out->Format.wFormatTag = WAVE_FORMAT_PCM;
	// set unused members to zero
	out->Format.cbSize = 0;
	out->Samples.wValidBitsPerSample = 0;
	out->dwChannelMask = 0;
	out->SubFormat = GUID_NULL;
}


int KSCore::OpenOutput(WAVEFORMATEXTENSIBLE *wfx)
{
	int n,num_attempt;
//	char errinfo[512];

	if (config->input_buffer_size != input_buffer_size) {
		InitInputBuffer(config->input_buffer_size);
	}
	if (config->output_buffer_size != packet_size) {
		InitOutputBuffer(config->output_buffer_size);
		InitPackets(FALSE);
	}


	OpenEnumerator();			// create filter list

	filter = PickFilter();	// pick filter to render to

	if(!filter)
	{
		MessageBox(hwnd, "Could not find suitable output filter", "Error!", MB_OK);
		return FALSE;
	}

	num_attempt = 0;

	DebugMsg("OpenOutput - Current Settings:\n"
				"Chosen Filter:       %s\n"
				"Input Buffer Size:   %d\n"
				"Output Buffer Size   %d\n"
				"Output Buffer Count: %d\n\n",
				filter->GetFriendlyName(),
				config->input_buffer_size,
				config->output_buffer_size,
				config->output_buffer_cnt);


	DebugMsg("Attempting to CreateRenderPin with the following WaveFormatExtensible (first attempt):\n"
				"Format.nChannels            = %d\n"
				"Format.nSamplesPerSec       = %d\n"
				"Format.wBitsPerSample       = %d\n"
				"Format.nBlockAlign          = %d\n"
				"Format.nAvgBytesPerSec      = %d\n"
				"Format.cbSize               = %d\n"
				"Samples.wValidBitsPerSample = %d\n"
				"dwChannelMask               = %08x\n\n",
				wfx->Format.nChannels,
				wfx->Format.nSamplesPerSec,
				wfx->Format.wBitsPerSample,
				wfx->Format.nBlockAlign,
				wfx->Format.nAvgBytesPerSec,
				wfx->Format.cbSize,
				wfx->Samples.wValidBitsPerSample,
				wfx->dwChannelMask);






	pin = filter->CreateRenderPin(&wfx->Format, FALSE);

	if (!pin)
	{
		DebugMsg("Failed to create pin on first attempt!\n");


		// driver can't handle WAVEFORMATEXTENSIBLE, so fall back to
		// WAVEFORMATEX format descriptor and try again

		ConvertWaveFormatExtensibleToWaveFormat(wfx,wfx);


		DebugMsg("Attempting to CreateRenderPin with the following WaveFormat (2nd attempt):\n"
					"Format.nChannels            = %d\n"
					"Format.nSamplesPerSec       = %d\n"
					"Format.wBitsPerSample       = %d\n"
					"Format.nBlockAlign          = %d\n"
					"Format.nAvgBytesPerSec      = %d\n"
					"Format.cbSize               = %d\n"
					"Samples.wValidBitsPerSample = %d\n"
					"dwChannelMask               = %08x\n\n",
					wfx->Format.nChannels,
					wfx->Format.nSamplesPerSec,
					wfx->Format.wBitsPerSample,
					wfx->Format.nBlockAlign,
					wfx->Format.nAvgBytesPerSec,
					wfx->Format.cbSize,
					wfx->Samples.wValidBitsPerSample,
					wfx->dwChannelMask);

		pin = filter->CreateRenderPin(&wfx->Format, FALSE);
		num_attempt++;
	}

	if (!pin)
	{
//		sprintf(errinfo,
//			"Unable to create output pin on %s attempt\n"
//			"Reason: %s\n"
//			"Format Info: nChannels %d Freq: %d Sample Size: %d",
//			((num_attempt) ? "second":"first"),
//			filter->GetError(),
//			wfx->Format.nChannels,
//			wfx->Format.nSamplesPerSec,
//			wfx->Format.wBitsPerSample
//			);
//
//		MessageBox(hwnd, errinfo, "Error!", MB_OK);
		return FALSE;
	}

	DebugMsg("OpenOutput: Successfully Created Pin, Format - BPS: %d FREQ: %d NCHANNELS: %d VALIDBPS: %d", wfx->Format.wBitsPerSample, wfx->Format.nSamplesPerSec, wfx->Format.nChannels, wfx->Samples.wValidBitsPerSample);

	// initialise packets and events ready for playback
	for (n = 0; n < num_packets; n++) {
		SetEvent(events[EV_KSPACKET+n]);
		packets[n].Header.FrameExtent = 0;	// empty packet
		packets[n].Header.DataUsed = 0;		// empty packet
	}
	ResetEvent(events[EV_MSG]);

	wfx_cur = *wfx;		// save format settings

	mode = MODE_STARTED;	// flag as started
	paused = FALSE;

	DebugMsg("Output Opened");

	return TRUE;
}


void KSCore::CloseOutput(void)
{
	// close the ouput pin
	if (pin) {
	   pin->ClosePin();
		pin = NULL;
	}

	// delete all filters
	CloseEnumerator();
	mode = MODE_STOPPED;
	DebugMsg("Output Closed");
}





void KSCore::InitPackets(int init_events)
{
	int n;

	for (n = 0; n < num_packets; n++) {
		if (init_events)
			events[EV_KSPACKET+n] = CreateEvent(NULL, TRUE, FALSE, NULL);
		memset(&packets[n],0,sizeof(DATA_PACKET));
		packets[n].Signal.hEvent = events[EV_KSPACKET+n];
		packets[n].Header.Data = output_buffer + n * packet_size;
		packets[n].Header.FrameExtent = 0;//packet_size;
		packets[n].Header.DataUsed    = 0;//packet_size;
		packets[n].Header.Size = sizeof(packets[n].Header);
		packets[n].Header.PresentationTime.Numerator = 1;
		packets[n].Header.PresentationTime.Denominator = 1;

	}
}

void KSCore::InitInputBuffer(int size)
{
	if (input_buffer) {
		free(input_buffer);
	}
	input_buffer = (char *)malloc(size);
	input_buffer_size = size;

	FlushInputBuffer();
	DebugMsg("input buffer configured to %d bytes",input_buffer_size);
}


void KSCore::InitOutputBuffer(int size)
{
	if (output_buffer) {
		free(output_buffer);
	}
	output_buffer = (char *)malloc(size * num_packets);
	packet_size = size;
	DebugMsg("output buffer configured to %d bytes per packet",packet_size);
}


KSCore::KSCore(HWND hwnd, HINSTANCE hinst, class Config *config)
{

	kscore = this;

	this->hwnd = hwnd;
	this->hinst = hinst;
	this->config = config;

	diagnostic_wnd = NULL;

	packet_size = 8192;
	the_enumerator = NULL;
	filter = NULL;
	pin = NULL;
	state = KSSTATE_STOP;
	volume = 255;
	num_packets = config->output_buffer_cnt;
	should_be_playing = FALSE;

	events = (HANDLE *)malloc(sizeof(HANDLE)*(EV_KSPACKET+num_packets));	// one event for each packet, plus other synch objects

	input_buffer = NULL;
	output_buffer = NULL;
	InitInputBuffer(config->input_buffer_size);
	InitOutputBuffer(config->output_buffer_size);
	InitPackets(TRUE);

	cur_packet = -1;													// start with first packet

	events[EV_MSG]    = CreateEvent(NULL, TRUE, FALSE, NULL);	// extra event for control messages
	sleep_event       = CreateEvent(NULL, TRUE, FALSE, NULL);	// extra event for control messages

	quit = FALSE;
	mode = MODE_STOPPED;

	InitializeCriticalSection(&cs_msg);
	InitializeCriticalSection(&ks_access);

	EnableLogging(config->logging_enabled);


	OpenEnumerator();

	player_thread = CreateThread(NULL,0,KSCore::PlayerThreadWrapper,(LPVOID)NULL,0,&threadid);
	SetThreadPriority(player_thread,THREAD_PRIORITY_TIME_CRITICAL);

	DebugMsg("Constructed KSCore (Input Size:%d) (Output Size %d) (Output Cnt %d)", config->input_buffer_size, config->output_buffer_size, config->output_buffer_cnt);
}




KSCore::~KSCore()
{
	int n;

	End(TRUE);	// ensure playback is stopped

	CloseDiagnosticWindow();

	// Get Maintenance thread to quit
	DebugMsg("Attempting to close player thread");
	quit = TRUE;						
	SetEvent(events[EV_MSG]);
	WakePlayer();							// ensure player is up and running
	while(player_thread) Sleep(10);	// wait for thread to self terminate
	DebugMsg("Player thread has closed");

	DeleteCriticalSection(&ks_access);
	DeleteCriticalSection(&cs_msg);

	CloseHandle(events[EV_MSG]);
	for (n = 0; n < num_packets; n++) {
		CloseHandle(events[EV_KSPACKET+n]);
	}
	CloseHandle(sleep_event);
	free(output_buffer);
	free(input_buffer);
	free(events);
	DebugMsg("Destructed KSCore");
}


int KSCore::Start(int samplerate, int numchannels, int bitspersamp)
{
	int rv;

	EnterCriticalSection(&ks_access);
	rv = StartInternal(samplerate,numchannels,bitspersamp);
	LeaveCriticalSection(&ks_access);

	return rv;
}


int KSCore::StartInternal(int samplerate, int numchannels, int bitspersamp)
{
	HRESULT hr;
	WAVEFORMATEXTENSIBLE wfx;
	unsigned int channel_masks[] = {
		0,
		SPEAKER_FRONT_CENTER,
		SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT,
		SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT|SPEAKER_FRONT_CENTER,
		SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT|SPEAKER_BACK_LEFT|SPEAKER_BACK_RIGHT,
		SPEAKER_FRONT_CENTER|SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT|SPEAKER_BACK_LEFT|SPEAKER_BACK_RIGHT,
		SPEAKER_FRONT_CENTER|SPEAKER_FRONT_LEFT|SPEAKER_FRONT_RIGHT|SPEAKER_BACK_LEFT|SPEAKER_BACK_RIGHT|SPEAKER_LOW_FREQUENCY
	};


	memset(&wfx,0,sizeof(wfx));
	wfx.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
	wfx.Format.nChannels = numchannels;
	wfx.Format.nSamplesPerSec = samplerate;
	wfx.Format.wBitsPerSample = (bitspersamp+7)&0xfffffff8;	// round up to nearest eight
	wfx.Format.nBlockAlign    = numchannels*wfx.Format.wBitsPerSample/8;	// I think
	wfx.Format.nAvgBytesPerSec = wfx.Format.nSamplesPerSec * wfx.Format.nBlockAlign;
	wfx.Format.cbSize = sizeof(WAVEFORMATEXTENSIBLE)-sizeof(WAVEFORMATEX);
	wfx.Samples.wValidBitsPerSample = bitspersamp;
	wfx.dwChannelMask = channel_masks[min(numchannels,6)];
	wfx.SubFormat = KSDATAFORMAT_SUBTYPE_PCM;


	//
	// 24 Bit Bodge
	// It appears the KS interface cannot handle 24 bit data "as is" but instead requires it packaged
	// in a 32 bit wrapper, presumably for alignment purposes.
	// So we need to special case data of this type and reformat it as appropriate
	//
	switch(bitspersamp)
	{
	case 20:
		MessageBox(hwnd, "20 bit audio is not currently supported", "Error!", MB_OK);
		return -1;

	case 24:
		if (numchannels == 2) {
			wfx.Format.wBitsPerSample = 32;
			wfx.Format.nBlockAlign    = 8;
		}
		else if (numchannels == 1) {
			wfx.Format.wBitsPerSample = 32;
			wfx.Format.nBlockAlign    = 4;
		}
		else {
			MessageBox(hwnd, "Only 1 or 2 channels is supported for 24 bit playback.", "Error!", MB_OK);
			return -1;
		}
		break;

	default:
		break;
	}



	// Gapless enhancement
	if (mode == MODE_STARTED) {	// already playing
		if (!paused) {					// if STARTED and paused, we may as well delete and rebuild the output filter
			if (config->gapless_enabled) {
				if (0 == memcmp(&wfx_cur,&wfx,sizeof(WAVEFORMATEXTENSIBLE))) {
					should_be_playing = TRUE;
					written_data = 0;					// reset data count
					return 1000;	// no format change detected, so just carry on
				}
				// may be running in WAVEFORMAT mode if WAVEFORMATEXTENSIBLE failed, lets check
				{
					WAVEFORMATEXTENSIBLE wf;
					ConvertWaveFormatExtensibleToWaveFormat(&wf,&wfx);
					if (0 == memcmp(&wfx_cur,&wf,sizeof(WAVEFORMATEXTENSIBLE))) {
						should_be_playing = TRUE;
						written_data = 0;					// reset data count
						return 1000;						// no format change detected, so just carry on
					}
				}


			}
		}
		DebugMsg("StartInternal - Forcing End");
		End(TRUE);	// format changed, or not gapless, so close existing filter
	}

	if (!OpenOutput(&wfx)) {
		return -1;
	}

	written_data = 0;

	// set new state
	state = KSSTATE_RUN;
	hr = pin->SetState(state);

	if(!SUCCEEDED(hr))
	{
		return -1;
	}


	should_be_playing = TRUE;
	return 1000;// return max latency //max_latency;
}



void KSCore::End(void)
{
	DebugMsg("End()");
	should_be_playing = FALSE;
	End(FALSE);
}

void KSCore::End(int force)
{
	EnterCriticalSection(&ks_access);
	DebugMsg("End(force=%d)",force);
	EndInternal(force);
	LeaveCriticalSection(&ks_access);
}

//
// Ends playback - force parameter causes function to ignores gapless mode
//
void KSCore::EndInternal(int force)
{
	DebugMsg("EndInternal() - mode=%d force=%d",mode,force);

	if (mode != MODE_STARTED)
		return;

	if (config->gapless_enabled && !force)
		return;

	if (paused) {
		// safe to do this here as player thread must be in thread_paused state
		// at this point because paused waits for that to happen before proceeding
		FlushInputBuffer();
	}

	mode = MODE_STOPPING;	// tell player thread that we are stopping

	// wait for player thread to empty the intermediate buffer
	while(input_buffer_len) Sleep(10);

	// track has ended naturally, so we need to allow all packets to play out
	// otherwise we'll truncate it
	if (!paused)
		WaitForMultipleObjects(num_packets, events+EV_KSPACKET, TRUE, INFINITE);

	// stop the filter - it will be getting starved anyway
	if (pin) {
		HRESULT hr;
		hr = pin->SetState(KSSTATE_PAUSE);
		hr = pin->SetState(KSSTATE_STOP);
	}

	// need to be sure all packets have been released before we clean up
	// otherwise - crash city arizona!
	WaitForMultipleObjects(num_packets, events+EV_KSPACKET, TRUE, INFINITE);

	CloseOutput();

}


int KSCore::SetPause(int pause_state)
{
	int rv;
	HRESULT hr;

	EnterCriticalSection(&ks_access);

	rv = paused;

	if (mode == MODE_STARTED && pin) {
		paused = pause_state;
		if (paused) {
			while(!thread_paused) Sleep(10);
			state = KSSTATE_PAUSE;
			hr = pin->SetState(state);
		}
		else {
			state = KSSTATE_RUN;
			hr = pin->SetState(state);
//			Sleep(2000);							// need to give thread inside KS filter time to wake before we wake the player thread and start sending packets to it
			WakePlayer();
		}
	}
	LeaveCriticalSection(&ks_access);
	return rv;
}



int KSCore::MemCpyWithUnpack(int *out, char *in, int in_len)
{
	int out_len;
	union {
		int i;
		char c[4];
	} sample;


	out_len = 0;

	while(in_len > 0) {
		sample.c[0] = 0;
		sample.c[1] = in[0];
		sample.c[2] = in[1];
		sample.c[3] = in[2];
		*out++ = sample.i;
		out_len += 4;
		in += 3;
		in_len -= 3;
	}

	return out_len;
}


int KSCore::WritePacket(char *buf, int len)
{
	int amt_free;

	if (mode != MODE_STARTED)
		return 1;

	if (!len)
		return 0;

//	WakePlayer();

	amt_free = input_buffer_size - input_buffer_len;


	if (wfx_cur.Samples.wValidBitsPerSample == 24) {
		int acnt_24,bcnt_24;
		int acnt_32,bcnt_32;
		int len_32;
		
		if (amt_free >= (len/3)*4) { // enough space?
			acnt_32 = input_buffer_size - (write_ptr-input_buffer);		// how much to end of buffer?
			len_32 = (len / 3) * 4;										// get count when expanded from 24 bit to 32
			acnt_32 = min(acnt_32,len_32);							// size of first half (expanded)
			bcnt_32 = 0;
			if (acnt_32 < len_32) {										// packet straddles buffer end
				bcnt_32 = len_32 - acnt_32;							// size of second half (expanded)
			}
			acnt_24 = (acnt_32 / 4) * 3;
			bcnt_24 = (bcnt_32 / 4) * 3;
			if (acnt_24) {
				MemCpyWithUnpack((int *)write_ptr,buf,acnt_24);
				write_ptr += acnt_32;
			}
			if (bcnt_24) {						// deal with straddled write
				write_ptr = input_buffer;
				MemCpyWithUnpack((int *)input_buffer,buf+acnt_24,bcnt_24);
				write_ptr += bcnt_32;
			}
			input_buffer_len += len_32;		// update size of data in buffer
			WakePlayer();
			return 0;							// signal success
		}
	}
	else {
		int acnt,bcnt;

		if (amt_free >= len) { // enough space?
			acnt = input_buffer_size - (write_ptr-input_buffer);		// how much to end of buffer?
			acnt = min(acnt,len);									// size of first half
			bcnt = 0;
			if (acnt < len) {											// packet straddles buffer end
				bcnt = len - acnt;
			}
			if (acnt) {
				memcpy((void *)write_ptr,buf,acnt);
				write_ptr += acnt;
			}
			if (bcnt) {						// deal with straddled write
				write_ptr = input_buffer;
				memcpy(input_buffer,buf+acnt,bcnt);
				write_ptr += bcnt;
			}
			input_buffer_len += len;		// update size of data in buffer
			WakePlayer();
			return 0;						// signal success
		}
	}

	return 1;							// signal failure
}



int KSCore::CanWrite(void)
{
	int effective_packet_size,effective_amt_free,rv;
	int amt_free;

	if (mode != MODE_STARTED)
		return 0;

	amt_free = input_buffer_size - input_buffer_len;
	if (wfx_cur.Samples.wValidBitsPerSample == 24) {
		effective_packet_size = (packet_size / 3) * 4;
		effective_amt_free = (amt_free / 4) * 3;
	}
	else {
		effective_packet_size = packet_size;
		effective_amt_free = amt_free;
	}


	if (amt_free >= effective_packet_size)
		rv = effective_amt_free;//packet_size;
	else
		rv = 0;

	return rv;
}


int KSCore::IsPlaying(void)
{
	DWORD rv;
	
	if (mode != MODE_STARTED)
		return 0;

	if (input_buffer_len)	// something in buffer, so, still playing
		return 0;

	rv = WaitForMultipleObjects(num_packets, events+EV_KSPACKET, TRUE, 0);
	if (rv == WAIT_TIMEOUT)
		return 0;
	else
		return 1;
}


int KSCore::GetWrittenTime(void)
{
	unsigned int bytes_per_sec, t;
	bytes_per_sec = wfx_cur.Format.nAvgBytesPerSec;

		
	t =  written_data / bytes_per_sec * 1000;							// time to nearest second in mS
	t += (written_data % bytes_per_sec) * 1000 / bytes_per_sec;	// add in fractional component

	return t;
}

void KSCore::SetVolume(int vol)
{
	if (vol >= 0)		// seem to get a negative value passed through immediately after playback begins
		volume = max(0,min(255,vol));
}



static int gdiDrawRect(HDC hdc, int left, int top, int right, int bottom, HBRUSH br)
{
	RECT trect;

	if (left >= right)
		return (FALSE);

	if (bottom <= top)
		return (FALSE);

	trect.left = left;
	trect.right = right;
	trect.top = top;
	trect.bottom = bottom;

	FillRect(hdc, &trect, br);

	return (TRUE);
}


static void gdiPrint(HDC hdc, int x, int y, char *txt,...)
{
	va_list args;
	char buff[256];

	va_start(args,txt);
	vsprintf(buff, txt, args);
	va_end(args);

	TextOut(hdc, x, y, buff, strlen(buff));
}

int gdiPrintLen(HDC hdc, char *txt,...)
{
	va_list args;
	char buff[256];
	SIZE size;

	va_start(args,txt);
	vsprintf(buff, txt, args);
	va_end(args);

	GetTextExtentPoint32(hdc, buff, strlen(buff), &size);
	return (size.cx);
}

int gdiPrintHeight(HDC hdc, char *txt,...)
{
	va_list args;
	char buff[256];
	SIZE size;

	va_start(args,txt);
	vsprintf(buff, txt, args);
	va_end(args);

	GetTextExtentPoint32(hdc, buff, strlen(buff), &size);
	return (size.cy);
}

void KSCore::DrawDiagnostics(HWND hwnd, HDC hdc)
{
	HDC cdc;
	HBITMAP ccb;
	HBRUSH brsh,a_brsh,b_brsh;
	RECT r;
	int y,ysize;
	int n;
	LOGBRUSH lb;

	GetClientRect(hwnd,&r);
	cdc = CreateCompatibleDC(hdc);
	if (cdc) {
		ccb = CreateCompatibleBitmap(hdc,r.right-r.left,r.bottom-r.top);
		if (ccb) {
			int s,v,l,w;
			float lat = 0.0f;

			y = 20;
			ysize = 20;

			SelectObject(cdc,ccb);
			lb.lbStyle = BS_SOLID;
			lb.lbColor = GetSysColor(COLOR_BTNFACE);
			brsh    = CreateBrushIndirect(&lb);
			FillRect(cdc,&r,brsh);
			DeleteObject(brsh);

			lb.lbStyle = BS_SOLID;
			lb.lbColor = 0x0000ff00;
			a_brsh    = CreateBrushIndirect(&lb);
			lb.lbStyle = BS_SOLID;
			lb.lbColor = 0xa0a0a0a0;
			b_brsh    = CreateBrushIndirect(&lb);

			SetBkMode(cdc,TRANSPARENT);


			gdiPrint(cdc,20,y,"Player Thread State: %s", dbg_player_sleeps ? "SLEEPING" : "AWAKE");
			y += ysize<<1;

		 	gdiPrint(cdc,20,y,"Input Buffer:");
			y += ysize;

			s = input_buffer_size;
			if (s && wfx_cur.Format.nAvgBytesPerSec) {
				lat = (float)s / (float)wfx_cur.Format.nAvgBytesPerSec;
			}
		 	gdiPrint(cdc,20,y,"Size: %d bytes (%.1fmS)", s,lat * 1000.0f);
			y += ysize;
			if (s) {
			 	gdiPrint(cdc,20,y,"Usage:");
			 	l = gdiPrintLen(cdc,"Usage:") + 10;
				v = input_buffer_len * 100 / s;
				gdiDrawRect(cdc,20+l,y,20+l+200,y+ysize,b_brsh);
				gdiDrawRect(cdc,20+l,y,20+l+(v<<1),y+ysize,a_brsh);
			 	w = gdiPrintLen(cdc,"%d%%", v);
			 	gdiPrint(cdc,20+l+100-(w>>1),y,"%d%%", v);
			}
			y += ysize<<1;


			gdiPrint(cdc,20,y,"Output Buffers:");
			y += ysize;

			if (packet_size && wfx_cur.Format.nAvgBytesPerSec) {
				lat = (float)packet_size / (float)wfx_cur.Format.nAvgBytesPerSec;
			}
			gdiPrint(cdc,20,y,"Size: %d bytes (%.1fmS x %d)", packet_size, lat * 1000.0f,num_packets);
			y += ysize;

			for (n = 0; n < num_packets; n++) {
				gdiPrint(cdc,20,y,"%d : %s", n, (WaitForSingleObject(events[n+EV_KSPACKET], 0) == WAIT_OBJECT_0) ? "FREE" : "IN USE");
				y += ysize;
			}				

			
			y += ysize;
			gdiPrint(cdc,20,y,"Output Format:");
			y += ysize;
			gdiPrint(cdc,20,y,"Frequency: %dHz", wfx_cur.Format.nSamplesPerSec);
			y += ysize;
			if (wfx_cur.Samples.wValidBitsPerSample)
				gdiPrint(cdc,20,y,"Sample Size: %d bits", wfx_cur.Samples.wValidBitsPerSample);
			else
				gdiPrint(cdc,20,y,"Sample Size: %d bits", wfx_cur.Format.wBitsPerSample);
			y += ysize;
			gdiPrint(cdc,20,y,"Number of Channels: %d", wfx_cur.Format.nChannels);
			y += ysize;

			BitBlt(hdc,0,0,r.right-r.left,r.bottom-r.top,cdc,0,0,SRCCOPY);
			DeleteObject(ccb);
			DeleteObject(a_brsh);
			DeleteObject(b_brsh);
		}
		DeleteObject(cdc);
	}

	
}


VOID CALLBACK KSCore::DiagnosticRefresh(HWND hwnd, UINT msg, UINT_PTR id, DWORD dwTime)
{
	if (id == 1) {
		InvalidateRect(diagnostic_wnd,NULL,FALSE);
	}
}

VOID CALLBACK KSCore::DiagnosticRefreshWrapper(HWND hwnd, UINT msg, UINT_PTR idevent, DWORD dwTime)
{
	kscore->DiagnosticRefresh(hwnd, msg, idevent, dwTime);
}

BOOL CALLBACK KSCore::DiagnosticDialog(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	HDC hdc;
	PAINTSTRUCT ps;

	switch(msg)
	{
	case WM_CREATE:
		break;

	case WM_SHOWWINDOW:
		SetTimer(hwnd, 1, 20, DiagnosticRefreshWrapper);
		break;

	case WM_CLOSE:
			DestroyWindow(hwnd);
			break;

	case WM_DESTROY:
		if (diagnostic_wnd) {
			KillTimer(hwnd,1);
			diagnostic_wnd = NULL;
		}
		break;

	case WM_COMMAND:
		switch(LOWORD(wparam))
		{
		case IDCANCEL:
			EndDialog(hwnd,0);
			break;

		default:
			break;
		}
		break;

	case WM_PAINT:
		hdc = BeginPaint(hwnd,&ps);
		DrawDiagnostics(hwnd,hdc);
		EndPaint(hwnd, &ps);
		break;

	default:
		break;
	}

	return (FALSE);
}

BOOL CALLBACK KSCore::DiagnosticDialogWrapper(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	return kscore->DiagnosticDialog(hwnd, msg, wparam, lparam);
}


void KSCore::OpenDiagnosticWindow(void)
{
	DebugMsg("OpenDiagnosticWindow - current thread (%d)",GetCurrentThread());
	if (!diagnostic_wnd) {
		diagnostic_wnd = CreateDialog(hinst,MAKEINTRESOURCE(IDD_DIAGNOSTIC),NULL, DiagnosticDialogWrapper);
		if (diagnostic_wnd) {
			ShowWindow(diagnostic_wnd,SW_RESTORE);
		}
	}
}

void KSCore::CloseDiagnosticWindow(void)
{
	HWND wnd;

	wnd = diagnostic_wnd;

	if (wnd) {
		DestroyWindow(wnd);
		diagnostic_wnd = NULL;
	}
}
