/* ==================================================================
   File:        main.cpp
   Author:      Steve Monks
   Description:
	WinAmp Kernel Streaming plugin interface

 	$Id: out_ks.cpp,v 1.12 2007/01/06 21:50:15 Steve Exp $
	$Log: out_ks.cpp,v $
	Revision 1.12  2007/01/06 21:50:15  Steve
	Changes for version 3.61
	Fixed issue where log files could be created in different locations depending
	on how Winamp was launched.
	
	Revision 1.11  2006/04/09 19:51:53  Steve
	Version 3.6
	Added logarithmic volume control
	Added diagnostic logging
	
	Revision 1.10  2006/03/12 15:57:09  Steve
	Fixed incorrect play reporting issue
	
	Revision 1.9  2006/02/11 14:06:00  Steve
	Fix for Gapless Mode crashes when format changes.
	Fix for Gapless Mode lockup when track changed while paused
	Added format info to status window
	Fixes to Release CFG project setting (removed debug info and _DEBUG define)
	
	Revision 1.8  2006/02/04 21:57:38  Steve
	Added out shutdown of KS when buffer underruns in gapless mode
	tweaks to readme
	
	Revision 1.7  2006/02/04 20:59:44  Steve
	Initial import of readme.html
	
	Added gapless mode (for evaluation purposes)
	
	Revision 1.6  2006/01/30 21:07:29  Steve
	Added protection against daft values to config stuff
	Fixed config dialog so that KSCore is only recreated when settings actually change
	Changed buffer size readouts to KB in config dialog
	Fixed potential problem with lost volume settings when KScore is deleted and recreated
	
	Revision 1.5  2006/01/29 21:33:57  Steve
	Added an adjustment for the number of output buffers
	Removed buffer size radio buttons and replaced with sliders
	Now deletes and recreates the KSCore object when a new track starts after changing the settings
	
	Revision 1.4  2006/01/08 22:44:38  Steve
	Added status window
	Added increased buffer settings
	Fixed .d00 file problems (returned truthful buffer sizes rather than 8192)
	
	Revision 1.3  2006/01/06 22:26:20  Steve
	Fixed problem with Winamp Music and .mod playback (checking against packet_size
	rather than len inside Write function).
	Removed delay before player thread sleeps when starved of data
	
	Revision 1.2  2006/01/03 21:46:13  Steve
	Initial import of V3.0.
	
	This is a complete rewrite of the plugin and Kernel streaming code.
	The only code retained from the previous 2.x version is the Microsoft
	sample code and the filter enumeration.
	
	
================================================================== */

#include "kssample.h"
#include "out.h"
#include "resource.h"
#include "stdarg.h"
#include "error.h"
#include "tooltips.h"
#include "config.h"
#include "deviceinfo.h"
#include "kscore.h"


#define SQUARE(a) ((a) * (a))

#define PI_VER2 "v3.61"

#define PI_VER PI_VER2 " (x86)"

volatile int start_offset;

static TTItem *ttitem = NULL;
static DeviceInfoItem *device_info_list = NULL;
static Config *config;
static KSCore *kscore = NULL;
static int settings_changed = FALSE;
static int plugin_vol = 255;					// last (valid) volume set by winamp, used to restore volume if we need to recreate the KSCore object
extern Out_Module out;
HWND winamp_wnd = NULL;

static int log_volume_table[256];			// logarithmic volume table

static int input_buffer_sizes[] = {
	16  * 1024,
	32  * 1024,
	64  * 1024,
	128 * 1024
};

#define MAX_INPUT_SIZES (sizeof(input_buffer_sizes) / sizeof(int))


static int output_buffer_sizes[] = {
	8  * 1024,
	16 * 1024,
	32 * 1024
};

#define MAX_OUTPUT_SIZES (sizeof(output_buffer_sizes) / sizeof(int))


#define MAX_OUTPUT_CNT (8)


static int BufferSizeToIndex(int *buffer, int range, int size)
{
	int n;

	for (n = 0; n < range; n++) {
		if (size == buffer[n])
			return n;
	}
	// unknown value
	return 0;
}


#define InputBufferSizeToIndex(a)  BufferSizeToIndex(input_buffer_sizes,  MAX_INPUT_SIZES,  (a))
#define OutputBufferSizeToIndex(a) BufferSizeToIndex(output_buffer_sizes, MAX_OUTPUT_SIZES, (a))


BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH) DisableThreadLibraryCalls((HMODULE)hInst);
	return TRUE;
}


// Sets KSCore's volume, applies log table is required and keeps track of requested volume in case
// it needs to be reapplied for some reason

static void SetPluginVolume(int volume)
{
	plugin_vol = volume;					// keep track of current volume in case we recreate the kscore object

	if (kscore) {
		if (config->logarithmic_volume) {
			volume = log_volume_table[volume];
		}
		kscore->SetVolume(volume);
	}
}




//
// The Configuration dialog
//

static BOOL CALLBACK ConfigDialog(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	static int input_size;
	static int output_size;
	static int output_cnt;
	int vol_enabled;
	int gapless_enabled;
	int logarithmic_volume;
	int logging_enabled;

	char ascnum[16];
	int size_ix;
		
	switch(msg)
	{
	case WM_CREATE:
		return TRUE;

	case WM_DESTROY:
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_OUTPUTLIST));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_ENABLEVOLUMECONTROL));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_GAPLESSMODE));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_DEFAULTS));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDOK));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDCANCEL));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_LOGARITHMICVOLUMECONTROL));
			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_ENABLELOGGING));

//			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_INPUT16K ));
//			ttitem->RemoveTool(GetDlgItem(hwnd,IDC_OUTPUT4K ));
														  
		return FALSE;

	case WM_SHOWWINDOW:

		{
			// setup tooltips
			ttitem->AddTool(GetDlgItem(hwnd,IDC_OUTPUTLIST),					     	"Choose the output device you wish to stream to");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_ENABLEVOLUMECONTROL),		     	"If checked the WinAmp volume control will work although, if outputting to SPDIF having the volume set to any value other than 100% will cause the bitstream to be modified");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_GAPLESSMODE),					     	"If checked, playback will not be interrupted between tracks");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_DEFAULTS), 						     	"Restores the default configuration");
			ttitem->AddTool(GetDlgItem(hwnd,IDOK),					     					"Accepts and applies the new settings");
			ttitem->AddTool(GetDlgItem(hwnd,IDCANCEL),			     					"Disregards the new settings");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_LOGARITHMICVOLUMECONTROL),			"Makes it easier to finely adjust low volume levels");
			ttitem->AddTool(GetDlgItem(hwnd,IDC_ENABLELOGGING),     					"Activates diagnostic logging. Diagnostic information is output to \"kernel streaming.txt\"");

//			ttitem->AddTool(GetDlgItem(hwnd,IDC_INPUT16K ),		     					"Sets the size of the input buffer. Smaller values are more responsive, bigger values are less prone to breakup");
//			ttitem->AddTool(GetDlgItem(hwnd,IDC_OUTPUT4K ),		     					"Sets the size of the output buffer. Smaller values are more responsive, bigger values are less prone to breakup");
		}


		{
			DeviceInfoItem *i;
			int ix;

			i = device_info_list;

			while(i) {
				ix = SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_ADDSTRING, 0, (LPARAM)i->GetFriendlyName());			// add name to list box
				if (!config->out_name[0] && !ix)
					SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETCURSEL, ix, 0);												// highlight first entry if no current setup
				else if (0 == strcmp(i->GetDevicePath(),config->out_name))
					SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETCURSEL, ix, 0);												// highlight if current
				SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_SETITEMDATA, ix, (LPARAM)i->GetDevicePath());				// add device path to list box
				i = i->Next();
			}
		}

		{
			int sw,sh;
			RECT r;

			sw = GetSystemMetrics(SM_CXSCREEN);
			sh = GetSystemMetrics(SM_CYSCREEN);
			if (GetWindowRect(hwnd,&r)) {
				SetWindowPos(hwnd,HWND_TOP,(sw-(r.right-r.left))>>1,(sh-(r.bottom-r.top))>>1,0,0,SWP_NOSIZE);
			}
		}

		SendDlgItemMessage(hwnd,IDC_ENABLEVOLUMECONTROL,     BM_SETCHECK,(WPARAM)           (config->volume_enabled)      ? 1:0, 0);
		SendDlgItemMessage(hwnd,IDC_GAPLESSMODE,             BM_SETCHECK,(WPARAM)           (config->gapless_enabled)     ? 1:0, 0);
		SendDlgItemMessage(hwnd,IDC_LOGARITHMICVOLUMECONTROL,BM_SETCHECK,(WPARAM)           (config->logarithmic_volume)  ? 1:0, 0);
		SendDlgItemMessage(hwnd,IDC_ENABLELOGGING,           BM_SETCHECK,(WPARAM)           (config->logging_enabled)     ? 1:0, 0);

//		// input buffer
		input_size = config->input_buffer_size;
		SendDlgItemMessage(hwnd,IDC_SL_INPUTSIZE, TBM_SETRANGE, FALSE, MAKELONG(0, MAX_INPUT_SIZES-1));
		SendDlgItemMessage(hwnd,IDC_SL_INPUTSIZE, TBM_SETPOS,   TRUE, InputBufferSizeToIndex(input_size));
		sprintf(ascnum,"%dKB",input_size/1024);
		SetDlgItemText(hwnd,IDC_DISP_INPUTSIZE, ascnum);

		// output buffer
		output_size = config->output_buffer_size;
		SendDlgItemMessage(hwnd,IDC_SL_OUTPUTSIZE, TBM_SETRANGE, FALSE, MAKELONG(0, MAX_OUTPUT_SIZES-1));
		SendDlgItemMessage(hwnd,IDC_SL_OUTPUTSIZE, TBM_SETPOS,   TRUE, OutputBufferSizeToIndex(output_size));
		sprintf(ascnum,"%dKB",output_size/1024);
		SetDlgItemText(hwnd,IDC_DISP_OUTPUTSIZE, ascnum);

		output_cnt = config->output_buffer_cnt;
		SendDlgItemMessage(hwnd,IDC_SL_OUTPUTCNT, TBM_SETRANGE, FALSE, MAKELONG(2, MAX_OUTPUT_CNT));
		SendDlgItemMessage(hwnd,IDC_SL_OUTPUTCNT, TBM_SETPOS,   TRUE, output_cnt);
		sprintf(ascnum,"%d",output_cnt);
		SetDlgItemText(hwnd,IDC_DISP_OUTPUTCNT, ascnum);

		return TRUE;

	case WM_HSCROLL:
		if (lparam == (LPARAM)GetDlgItem(hwnd,IDC_SL_INPUTSIZE)) {

			switch(LOWORD(wparam))
			{
			case SB_THUMBPOSITION:
			case TB_THUMBTRACK:
			case TB_LINEDOWN:
			case TB_LINEUP:
			case TB_BOTTOM:
			case TB_ENDTRACK:
			case TB_PAGEDOWN:
			case TB_PAGEUP:
			case TB_TOP:
				size_ix = SendDlgItemMessage(hwnd,IDC_SL_INPUTSIZE,TBM_GETPOS,0,0);
				size_ix = max(0,min(MAX_INPUT_SIZES-1,size_ix));	// clamp index (shouldn't be possible for index to be out of range, but, just in case)
				input_size = input_buffer_sizes[size_ix];
				sprintf(ascnum,"%dKB",input_size/1024);
				SetDlgItemText(hwnd,IDC_DISP_INPUTSIZE, ascnum);
				break;

			default:
				DebugMsg("Unknown message");
				break;
			}
			
		}
		else if (lparam == (LPARAM)GetDlgItem(hwnd,IDC_SL_OUTPUTSIZE)) {

			switch(LOWORD(wparam))
			{
			case SB_THUMBPOSITION:
			case TB_THUMBTRACK:
			case TB_LINEDOWN:
			case TB_LINEUP:
			case TB_BOTTOM:
			case TB_ENDTRACK:
			case TB_PAGEDOWN:
			case TB_PAGEUP:
			case TB_TOP:
				size_ix = SendDlgItemMessage(hwnd,IDC_SL_OUTPUTSIZE,TBM_GETPOS,0,0);
				size_ix = max(0,min(MAX_OUTPUT_SIZES-1,size_ix));	// clamp index (shouldn't be possible for index to be out of range, but, just in case)
				output_size = output_buffer_sizes[size_ix];
				sprintf(ascnum,"%dKB",output_size/1024);
				SetDlgItemText(hwnd,IDC_DISP_OUTPUTSIZE, ascnum);
				break;

			default:
				DebugMsg("Unknown message");
				break;
			}
			
		}
		else if (lparam == (LPARAM)GetDlgItem(hwnd,IDC_SL_OUTPUTCNT)) {

			switch(LOWORD(wparam))
			{
			case SB_THUMBPOSITION:
			case TB_THUMBTRACK:
			case TB_LINEDOWN:
			case TB_LINEUP:
			case TB_BOTTOM:
			case TB_ENDTRACK:
			case TB_PAGEDOWN:
			case TB_PAGEUP:
			case TB_TOP:
				output_cnt = SendDlgItemMessage(hwnd,IDC_SL_OUTPUTCNT,TBM_GETPOS,0,0);
				sprintf(ascnum,"%d",output_cnt);
				SetDlgItemText(hwnd,IDC_DISP_OUTPUTCNT, ascnum);
				break;

			default:
				DebugMsg("Unknown message");
				break;
			}
			
		}
		return TRUE;

	case WM_COMMAND:
		switch(LOWORD(wparam))
		{
		case IDCANCEL:
			EndDialog(hwnd,0);
			return TRUE;

		case IDOK:
			{
				int ix;
				ix = SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_GETCURSEL,0,0);
				if (ix != LB_ERR) {
					char *dev_path;
					dev_path = (char *)SendDlgItemMessage(hwnd,IDC_OUTPUTLIST,LB_GETITEMDATA,ix,0);						// retrieve output name
					if (dev_path) {
						if (strcmp(config->out_name,dev_path)) {
							strcpy(config->out_name,dev_path);
							settings_changed = TRUE;
						}
					}
				}

				if (input_size <= output_size) {
					MessageBox(hwnd, "The input buffer MUST be larger than output buffer.\nThis has been corrected.", "Warning!", MB_OK);
					input_size = output_size * 2;
				}

				vol_enabled        = SendDlgItemMessage(hwnd,IDC_ENABLEVOLUMECONTROL,     BM_GETSTATE,0,0) & BST_CHECKED;
				gapless_enabled    = SendDlgItemMessage(hwnd,IDC_GAPLESSMODE,             BM_GETSTATE,0,0) & BST_CHECKED;
				logarithmic_volume = SendDlgItemMessage(hwnd,IDC_LOGARITHMICVOLUMECONTROL,BM_GETSTATE,0,0) & BST_CHECKED;
				logging_enabled    = SendDlgItemMessage(hwnd,IDC_ENABLELOGGING,           BM_GETSTATE,0,0) & BST_CHECKED;

				if (config->volume_enabled != vol_enabled) {
					settings_changed = TRUE;
					config->volume_enabled = vol_enabled;
				}
				if (config->gapless_enabled != gapless_enabled) {
					settings_changed = TRUE;
					config->gapless_enabled = gapless_enabled;
				}
				if (config->input_buffer_size != input_size) {
					config->input_buffer_size = input_size;
					settings_changed = TRUE;
				}
				if (config->output_buffer_size != output_size) {
					config->output_buffer_size = output_size;
					settings_changed = TRUE;
				}
				if (config->output_buffer_cnt != output_cnt) {
					config->output_buffer_cnt = output_cnt;
					settings_changed = TRUE;
				}
				if (config->logarithmic_volume != logarithmic_volume) {
					config->logarithmic_volume = logarithmic_volume;
					SetPluginVolume(plugin_vol);
					settings_changed = TRUE;
				}
				if (config->logging_enabled != logging_enabled) {
					config->logging_enabled = logging_enabled;
					EnableLogging(config->logging_enabled);
					settings_changed = TRUE;
				}
  				config->Write();
			}
			EndDialog(hwnd,0);
			return (TRUE);

		case IDC_OPENSTATUS:
			kscore->OpenDiagnosticWindow();
			return TRUE;

		case IDC_DEFAULTS:
			config->Defaults();
			config->Write();
			settings_changed = TRUE;
			EndDialog(hwnd,0);
			return TRUE;

		default:
			return FALSE;
		}
	default:
		return FALSE;
	}
}


//
// Function to launch the configuration dialog
//

static void BeginConfigDialog(HWND hwnd)
{
	DialogBox(out.hDllInstance,MAKEINTRESOURCE(IDD_CONFIG),hwnd,ConfigDialog);
}



static void InitLogVolumeTable(void)
{
	int n;
	float v;

	for (n = 0; n < 256; n++) {
		v = 255.0f * SQUARE(((float)n)/255.0f);
		log_volume_table[n] = (int)v;
	}
}



static void About(HWND hwnd)
{
	MessageBox(hwnd,
			"Kernel Streaming Plugin " PI_VER "\n"
			"Copyright (c) 2006 Steve Monks.\n\n"
			"Contains portions of code designed\n"
			"and written by Chun-Yu Shei\n\n"
			"Compiled on " __DATE__ "\n","About",MB_OK);
}




//
// Initialise the Plugin
//

static void Initialise()
{
	config = new Config();

	DebugMsgInit(out.hMainWindow);
	EnableLogging(config->logging_enabled);
	DebugMsg("Kernel Streaming Plugin "PI_VER" Diagnostic log\n");

	InitLogVolumeTable();

	// initialise core system
	kscore = new KSCore(out.hMainWindow, out.hDllInstance, config);
	device_info_list = BuildDeviceInfoList(kscore->GetEnumerator());
	ttitem = new TTItem(out.hDllInstance);

	winamp_wnd = FindWindow("Winamp v1.x",NULL);
}


//
// Close the plugin and clean up
//

static void Quit()
{
	if (ttitem)
		delete ttitem;
	ttitem = NULL;

	DestroyDeviceInfoList();
	if (kscore)
		delete kscore;
	kscore = NULL;

	DebugMsgClose();
}





//
// called whenever play is pressed in Winamp or a new track begins playing
//

int StartPlayback(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
	start_offset = 0;
	if (settings_changed) {
		if (kscore)
			delete kscore;

		kscore = new KSCore(out.hMainWindow, out.hDllInstance, config);
		SetPluginVolume(plugin_vol);

		settings_changed = FALSE;
	}

	return kscore->Start(samplerate,numchannels,bitspersamp);
}


//
// Called whenever stop is pressed in Winamp or we reach the end of the current track
//

static void EndPlayback()
{
//	int winamp_status=SendMessage(winamp_wnd,WM_USER, 0, 104);
//	DebugMsg("status %d",winamp_status);
	kscore->End();
}

static int WriteBuffer(char *buf, int len)
{
	return kscore->WritePacket(buf,len);
}

static int CanWrite()
{
	return kscore->CanWrite();
}

static int IsPlaying()
{
	return kscore->IsPlaying();
}

static int Pause(int pause)
{
	return kscore->SetPause(pause);
}

static void SetVolume(int volume)
{
	if (volume >= 0 && volume <= 255) {
		SetPluginVolume(volume);

	}
}



static void SetPan(int pan)
{
	// Not supported
}

static int GetWrittenTime()
{
	return start_offset + kscore->GetWrittenTime();
}

static void Flush(int t)
{
	int a;

	start_offset = 0;
	a = t - GetWrittenTime();
	start_offset=a;
}
	


Out_Module out = {
	OUT_VER,
	"Kernel Streaming Output "PI_VER2,
	0x10000,	// was 33, but this is the ID of the disk writer, changed, but streaming media still blocks this plugin
	0, 		// hmainwindow
	0, 		// hdllinstance
	BeginConfigDialog,
	About,
	Initialise,
	Quit,
	StartPlayback,
	EndPlayback,
	WriteBuffer,
	CanWrite,
	IsPlaying,
	Pause,
	SetVolume,
	SetPan,
	Flush,
	GetWrittenTime,
	GetWrittenTime
};

extern "C"
{

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
	return &out;
}

}
